#ifndef _GNU_SOURCE
#define _GNU_SOURCE
#endif
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <fcntl.h>
#include <string.h>
#include <ctype.h>
#include <locale.h>
#include <list>
#include <vector>
using namespace std;

#ifdef USE_WIDEC
#define _XOPEN_SOURCE_EXTENDED
#include <wctype.h>
#include <cursesw.h>
char *mb_locale = "ru_RU.KOI8-R",
     *utf_locale = "ru_RU.UTF-8";
#else
#include <curses.h>
#endif

extern "C" {
#include <libdis.h>
}
#include <libelf/libelf.h>
#include "disarm.h"
#include "disasm.h"

/*
  symbol 0x9b is somehow bad, isprint returns true, but
  when we output it, cursor moves to left.
*/

bool x86_mode = true;

struct hints {
  const char *hint[11];
};

list <hints> hints_stack;
int hint_offs[] = { 0, 1, 9, 17, 25, 33, 41, 49, 57, 65, 74 };

void set_keyhint( int n, const char *hint ) {
  attrset( COLOR_PAIR(1) );
  mvaddstr( LINES-1, hint_offs[n], hint );
  hints_stack.back().hint[n] = hint;
}

void clear_keyhints() {
  attrset( COLOR_PAIR(1) );
  for ( int i = 1; i <= 10; i++ ) {
  	hints_stack.back().hint[i] = "      ";
	mvaddstr( LINES-1, hint_offs[i], hints_stack.back().hint[i] );
  }
}

void push_keyhints() {
  hints_stack.push_back( hints_stack.back() );
}

void pop_keyhints() {
  hints_stack.pop_back();
  attrset( COLOR_PAIR(1) );
  for ( int i = 1; i <= 10; i++ )
	mvaddstr( LINES-1, hint_offs[i], hints_stack.back().hint[i] );
}

void init_keyhints() {
  hints nhints;

  attrset( COLOR_PAIR(3) );
  mvaddstr( LINES-1, 0, "1" );
  mvaddstr( LINES-1, 7, " 2" );
  mvaddstr( LINES-1, 15, " 3" );
  mvaddstr( LINES-1, 23, " 4" );
  mvaddstr( LINES-1, 31, " 5" );
  mvaddstr( LINES-1, 39, " 6" );
  mvaddstr( LINES-1, 47, " 7" );
  mvaddstr( LINES-1, 55, " 8" );
  mvaddstr( LINES-1, 63, " 9" );
  mvaddstr( LINES-1, 71, " 10" );

  hints_stack.push_back( nhints );
  clear_keyhints();
}

void load_hiew_vmm();

void disasm_init( char *fname ) {
  int i;

  init_pair( 1, COLOR_BLACK, COLOR_CYAN );
  init_pair( 2, COLOR_CYAN, COLOR_BLUE );
  init_pair( 3, COLOR_WHITE, COLOR_BLACK );
  init_pair( 4, COLOR_BLACK, COLOR_WHITE );
  init_pair( 5, COLOR_YELLOW, COLOR_BLUE );
  init_pair( 6, COLOR_BLUE, COLOR_CYAN );
  move( 0, 0 );
  attrset( COLOR_PAIR(1) );
  for ( i = 0; i < COLS; i++ ) addch(' ');
  attrset( COLOR_PAIR(2) );
  for ( i = 0; i < (LINES-2)*COLS; i++ ) addch( ' ' );
  attrset( COLOR_PAIR(1) );
  for ( i = 0; i < COLS; i++ ) addch(' ');
  mvaddstr( 0, 1, fname );
  addch( ' ' );
  addch( ACS_VLINE );
  init_keyhints();
  set_keyhint( 1, "About" );
  set_keyhint( 3, "Edit" );
  set_keyhint( 4, "Mode" );
  set_keyhint( 5, "Goto" );
  set_keyhint( 6, "ObjTbl" );
  set_keyhint( 7, "Search" );
  set_keyhint( 10, "Quit" );
  load_hiew_vmm();
}

char buf[512];
int fd, buf_offs, fsize;

unsigned char *get_data( int offs ) {
  if ( offs < 0 || offs >= fsize ) return 0;
  if ( offs < buf_offs || offs > buf_offs + sizeof(buf) - 40 ) {
  	if ( offs - (int) sizeof(buf)/2 > 0 ) buf_offs = offs - sizeof(buf)/2;
	else buf_offs = 0;
	
  	lseek( fd, buf_offs, SEEK_SET );
	read( fd, buf, sizeof buf );
  }
  return (unsigned char*)buf + offs - buf_offs;
}

struct serv_desc_entry {
  int serv_num;
  char *serv_desc;
  vector <char*> func;
};

vector <serv_desc_entry> serv_list;

void load_hiew_vmm() {
  serv_desc_entry desc;
  char buf[80], *p, *serv_desc;
  FILE *f;
  int serv_num, rc;

  f = fopen( "/usr/local/lib/disasm/hiew.vmm", "r" );
  if ( !f ) {
  	f = fopen( "hiew.vmm", "r" );
	if ( !f ) return;
  }
  while ( !feof( f ) ) {
  	if ( (int)fgets( buf, sizeof buf, f ) == EOF ) break;
	rc = sscanf( buf, " [%04X.%a[^]]] ", &serv_num, &serv_desc );
	if ( rc == 2 ) {
		desc.serv_num = serv_num;
		desc.serv_desc = serv_desc;
		serv_list.push_back( desc );
		continue;
	}
// strip white space from start and end of string
	p = buf;
	while ( *p == ' ' || *p == '\r' || *p == '\n' ) p++;
	if ( p != buf ) memmove( buf, p, strlen(p) + 1 );
	p = buf + strlen(buf) - 1;
	while ( p > buf && (*p == ' ' || *p == '\r' || *p =='\n' )) p--;
	p[1] = 0;

	if ( *buf != ';' && strlen(buf) && serv_list.size() )
		serv_list.back().func.push_back( strdup(buf) );
  }
  fclose( f );
}

bool relop_callback_called;
int insn_offset;

void relative_op_callback( char *buf, int blen, int op ) {
  relop_callback_called = true;
  snprintf( buf, blen, "0x%08x", insn_offset + op );
}

int disassemble_address_wrapper_x86( int offs, char *buf, int buf_len, struct instr *insn ) {
  unsigned char *bp = (unsigned char*) buf;
  unsigned short *sp, service, function;
  int i, rc;

  if ( buf_len >= 6 && bp[0] == 0xcd && bp[1] == 0x20 ) {
  	sp = (unsigned short*)(buf + 2);
	function = sp[0];
	service = sp[1];
	if ( function & 0x8000 ) {
		strcpy( insn->mnemonic, "VxDjmp" );
		function &= ~0x8000;
	}
	else
		strcpy( insn->mnemonic, "VxDcall" );
	
	for ( i = 0; i < serv_list.size(); i++ )
		if ( serv_list[i].serv_num == service ) {
			if ( function < serv_list[i].func.size() )
				snprintf( insn->dest, 32, "%s.%s",
					serv_list[i].serv_desc,
					serv_list[i].func[function] );
			else
				snprintf( insn->dest, 32, "%s.%04X",
					serv_list[i].serv_desc, function );
			break;
		}
	if ( i == serv_list.size() )
		snprintf( insn->dest, 32, "%04X.%04X", service, function );
	insn->src[0] = 0;
	insn->aux[0] = 0;
	insn->size = 6;

	return 6;
  }
  insn_offset = offs;
  relop_callback_called = false;
  rc = disassemble_address( buf, buf_len, insn );
  if ( relop_callback_called ) {
  	insn_offset = offs + rc;
	rc = disassemble_address( buf, buf_len, insn );
  }
  return rc;
}

int disassemble_address_wrapper_arm( int offs, char *buf, int buf_len, struct instr *insn ) {
  unsigned int w;
  pInstruction instr;
  static char _buf[100];

  if ( offs & 3 ) return 0;

  w = *(unsigned int*)buf;
  instr = arm_instr_disassemble( w, offs, &arm_options );

  strcpy( _buf, instr->text );
  if ( instr->badbits || instr->oddbits ) {
	strcat( _buf, "  ;" );
	if ( instr->badbits ) strcat( _buf, " [illegal bits]" );
	if ( instr->oddbits ) strcat( _buf, " [unexpected bits]" );
  }

  strcpy( insn->mnemonic, _buf );
  insn->dest[0] = 0;
  insn->src[0] = 0;
  insn->aux[0] = 0;

  return 4;
}

int disassemble_address_wrapper( int offs, char *buf, int buf_len, struct instr *insn ) {
  if ( x86_mode )
	return disassemble_address_wrapper_x86( offs, buf, buf_len, insn );
  else
	return disassemble_address_wrapper_arm( offs, buf, buf_len, insn );
}

struct le_obj_entry {
  int vseg_size, rel_base, flags;
  int pm_index, pm_num;
  char name[5];
};

union obj_entry {
  le_obj_entry le;
  pe_obj_entry pe;
};

vector <obj_entry> obj_table;
int le_hdr_offs;

bool load_le_objtbl() {
  unsigned char *p;
  int hdr_offs, obj_offs, objnum, i, *pd;
  obj_entry ent;

  le_hdr_offs = -1;

  p = get_data( 0x3c );
  if ( !p ) return false;
  hdr_offs = *(int*)p;
  if ( hdr_offs < 0 || hdr_offs >= fsize ) return false;
  p = get_data( hdr_offs );
  if ( memcmp( p, "LE", 2 ) ) return false;
  p = get_data( hdr_offs + 0x40 );
  if ( !p ) return false;
  
  obj_offs = *(int*)p;
  objnum = *(int*)(p+4);
  for ( i = 0; i < objnum; i++ ) {
  	pd = (int*) get_data( hdr_offs + obj_offs + i * 0x18 );
	ent.le.vseg_size = pd[0];
	ent.le.rel_base = pd[1];
	ent.le.flags = pd[2];
	ent.le.pm_index = pd[3];
	ent.le.pm_num = pd[4];
	memcpy( ent.le.name, &pd[5], 4 );
	ent.le.name[4] = 0;
	obj_table.push_back( ent );
  }

  le_hdr_offs = hdr_offs;
  return true;
}

bool load_pe_objtbl() {
  int i, j, hdr_offs, objtbl_offs, obj_offs;
  unsigned char *p;
  pe_header hdr;
  obj_entry ent;

  for ( i = 0; i < 40; i++ ) {
  	if ( (i << 3) + 4 >= fsize ) return false;
  	p = get_data( i << 3 );
        if ( !memcmp( p, "PE\0\0", 4 ) ) break;
  }
  if ( i == 40 ) return false;

  hdr_offs = i << 3;
  for ( i = 0; i < sizeof hdr; i++ ) {
  	p = get_data( hdr_offs + i );
	if ( !p ) return false;
  	((unsigned char*)&hdr)[i] = *p;
  }

  objtbl_offs = hdr_offs + hdr.nthdr_size + 0x18;
  
  for ( i = 0; i < hdr.num_objects; i++ ) {
  	obj_offs = objtbl_offs + i * sizeof(pe_obj_entry);
  	for ( j = 0; j < sizeof(pe_obj_entry); j++ ) {
	  	p = get_data( obj_offs + j );
		if ( !p ) return false;
	  	((unsigned char*)&ent)[j] = *p;
	}
	ent.pe.rva += hdr.image_base;
	obj_table.push_back( ent );
  }

  return true;
}
  
bool load_elf_objtbl() {
  Elf *elf;
  int kind;

  elf_version( EV_CURRENT );
  elf = elf_begin( fd, ELF_C_READ, 0 );
  kind = elf_kind( elf );
  elf_end( elf );
  
  return kind == ELF_K_ELF;
}

struct select_list_ent {
  int x, y;
  char *str;
};

struct sel_list_state {
  WINDOW *w;
  vector <select_list_ent> lst;
  int attr1, attr2, lst_offs, ce, maxx, maxy;
};

inline void show_sel_list_entry( sel_list_state *s, int attr, int ent ) {
  wattrset( s->w, attr );
  mvwaddstr( s->w, s->lst[ent].y - s->lst_offs, s->lst[ent].x, s->lst[ent].str );
}

void display_sel_list( sel_list_state *s ) {
  int lst_offs = s->lst_offs;
  vector <select_list_ent> &lst = s->lst;
  int vsize = s->maxy - lst[0].y;

  for ( int i = 0; i < vsize && lst_offs + i < lst.size(); i++ ) {
  	if ( lst_offs + i == s->ce ) wattrset( s->w, s->attr1 );
	else wattrset( s->w, s->attr2 );
	mvwaddstr( s->w, lst[lst_offs + i].y - lst_offs,
		lst[lst_offs + i].x, lst[lst_offs + i].str );
  }

  if ( lst.size() > vsize ) {
  	int start, end;
  	start = lst_offs * vsize / lst.size();
	end = start + vsize * vsize / lst.size();
  	for ( int i = 0; i < s->maxy - lst[0].y; i++ ) {
		if ( i < start || i > end ) wattrset( s->w, s->attr2 );
		else wattrset( s->w, s->attr1 );
		mvwaddch( s->w, lst[lst_offs + i].y - lst_offs,
			s->maxx, ACS_CKBOARD );
	}
  }
}

int do_list_selection( WINDOW *w, int attr1, int attr2,
	vector <select_list_ent> &lst ) {
  int i, c;
  sel_list_state s;

  s.w = w;
  s.attr1 = attr1;
  s.attr2 = attr2;
  s.lst = lst;
  getmaxyx( w, s.maxy, s.maxx );
  s.maxy--;
  s.maxx--;
  s.ce = 0;
  s.lst_offs = 0;
  display_sel_list( &s );
  for (;;) {
  	wrefresh( w );
	move( 0, COLS-1 );
	refresh();
  	c = wgetch( w );
	if ( c == KEY_UP && s.ce != 0 ) {
		show_sel_list_entry( &s, attr2, s.ce );
		s.ce--;
		if ( s.ce < s.lst_offs ) {
			s.lst_offs--;
			display_sel_list( &s );
		}
		else show_sel_list_entry( &s, attr1, s.ce );
	}
	else if ( c == KEY_DOWN && s.ce < lst.size()-1 ) {
		show_sel_list_entry( &s, attr2, s.ce );
		s.ce++;
		if ( s.lst[s.ce].y - s.lst_offs > s.maxy - 1 ) {
			s.lst_offs++;
			display_sel_list( &s );
		}
		else show_sel_list_entry( &s, attr1, s.ce );
	}
	else if ( c == KEY_PPAGE ) {
		if ( s.ce != s.lst_offs ) {
			show_sel_list_entry( &s, attr2, s.ce );
			s.ce = s.lst_offs;
			show_sel_list_entry( &s, attr1, s.ce );
		}
		else {
			s.ce -= s.maxy - s.lst[0].y;
			if ( s.ce < 0 ) s.ce = 0;
			s.lst_offs = s.ce;
			display_sel_list( &s );
		}
	}
	else if ( c == KEY_NPAGE ) {
		int last_ent = s.lst_offs + s.maxy - s.lst[0].y - 1;
		if ( last_ent > s.lst.size() - 1 ) last_ent = s.lst.size()-1;
		if ( s.ce != last_ent ) {
			show_sel_list_entry( &s, attr2, s.ce );
			s.ce = last_ent;
			show_sel_list_entry( &s, attr1, s.ce );
		}
		else {
			s.ce += s.maxy - s.lst[0].y;
			if ( s.ce > s.lst.size() - 1 ) s.ce = s.lst.size()-1;
			s.lst_offs = s.ce - (s.maxy - s.lst[0].y - 1);
			if ( s.lst_offs < 0 ) s.lst_offs = 0;
			display_sel_list( &s );
		}
	}
	else if ( c == KEY_HOME ) {
		s.ce = 0;
		s.lst_offs = 0;
		display_sel_list( &s );
	}
	else if ( c == KEY_END && lst.size() ) {
		s.ce = lst.size()-1;
		s.lst_offs = s.ce - (s.maxy - s.lst[0].y - 1);
		if ( s.lst_offs < 0 ) s.lst_offs = 0;
		display_sel_list( &s );
	}
	else if ( c == '\n' ) return s.ce;
	else if ( c == 033 ) return -1;
  }
}

int (*view_objtbl)() = 0;

int view_le_objtbl() {
  WINDOW *w;
  int i, c, rc, offs;
  unsigned char *p;
  select_list_ent sle;
  vector <select_list_ent> lst;

  push_keyhints();
  clear_keyhints();

  w = newwin( 15, 51, (LINES-15)/2, (COLS-51)/2 );
  keypad( w, TRUE );
  box( w, ACS_VLINE, ACS_HLINE );
  wbkgd( w, COLOR_PAIR(6) );
  mvwprintw( w, 0, 2, "Object Table" );
  mvwaddstr( w, 1, 1, "Name RelBase  SegSize  PMIndex  PMNumber Flags" );
  for ( i = 0; i < obj_table.size(); i++ ) {
  	sle.x = 1;
	sle.y = i + 2;
	asprintf( &sle.str, "%4.4s %08x %08x %08x %08x %08x",
		obj_table[i].le.name, obj_table[i].le.rel_base,
		obj_table[i].le.vseg_size, obj_table[i].le.pm_index,
		obj_table[i].le.pm_num, obj_table[i].le.flags );
	lst.push_back( sle );
  }
  rc = do_list_selection( w, COLOR_PAIR(3), COLOR_PAIR(6), lst );
  if ( rc != -1 && lst.size() ) {
  	int dataoffs, pgidx, page, pgsize;

  	p = get_data( le_hdr_offs + 0x80 );
	dataoffs = *(int*)p;
	p = get_data( le_hdr_offs + 0x28 );
	pgsize = *(int*)p;
	p = get_data( le_hdr_offs + 0x48 );
	pgidx = le_hdr_offs + *(int*)p + (obj_table[rc].le.pm_index - 1) * 4;
	p = get_data( pgidx );
	page = *(unsigned short*)p + *(unsigned short*)(p + 2) - 1;
	offs = dataoffs + page * pgsize;
  }
  else offs = -1;

  delwin( w );
  pop_keyhints();

  return offs;
}

int view_pe_objtbl() {
  WINDOW *w;
  int i, c, rc, offs;
  unsigned char *p;
  select_list_ent sle;
  vector <select_list_ent> lst;

  push_keyhints();
  clear_keyhints();

  w = newwin( 15, 55, (LINES-15)/2, (COLS-55)/2 );
  keypad( w, TRUE );
  box( w, ACS_VLINE, ACS_HLINE );
  wbkgd( w, COLOR_PAIR(6) );
  mvwprintw( w, 0, 2, "Object Table" );
  mvwaddstr( w, 1, 1, "Name     VirtSize RVA      PhysSize PhysOffs Flags   " );
  for ( i = 0; i < obj_table.size(); i++ ) {
  	sle.x = 1;
	sle.y = i + 2;
	asprintf( &sle.str, "%-8.8s %08x %08x %08x %08x %08x",
		obj_table[i].pe.name, obj_table[i].pe.virtual_size,
		obj_table[i].pe.rva, obj_table[i].pe.physical_size,
		obj_table[i].pe.physical_offset, obj_table[i].pe.object_flags );
	lst.push_back( sle );
  }
  rc = do_list_selection( w, COLOR_PAIR(3), COLOR_PAIR(6), lst );

  if ( rc != -1 && lst.size() )
	offs = obj_table[rc].pe.physical_offset;
  else offs = -1;

  delwin( w );
  pop_keyhints();

  return offs;
}

int view_elf_objtbl() {
  WINDOW *w;
  int i, c, rc, offs;
  unsigned char *p;
  char *str;
  select_list_ent sle;
  vector <select_list_ent> lst;
  Elf *elf;
  Elf32_Ehdr *ehdr;
  Elf_Scn *scn;
  Elf32_Shdr *shdr;

  push_keyhints();
  clear_keyhints();

  w = newwin( 18, 61, (LINES-18)/2, (COLS-61)/2 );
  keypad( w, TRUE );
  box( w, ACS_VLINE, ACS_HLINE );
  wbkgd( w, COLOR_PAIR(6) );
  mvwprintw( w, 0, 2, "Object Table" );
  mvwaddstr( w, 1, 1, "Section name   Type     Address  FileOffs Size     Flags   " );

  elf = elf_begin( fd, ELF_C_READ, 0 );
  ehdr = elf32_getehdr( elf );
  scn = 0;
//  for ( i = 0; i < 15; i++ ) {
  for ( i = 0;; i++ ) {
  	scn = elf_nextscn( elf, scn );
	if ( !scn ) break;
	shdr = elf32_getshdr( scn );
	str = elf_strptr( elf, ehdr->e_shstrndx, shdr->sh_name );
  	sle.x = 1;
	sle.y = i + 2;
	asprintf( &sle.str, "%-14.14s %08x %08x %08x %08x %08x",
                str, shdr->sh_type, shdr->sh_addr, shdr->sh_offset,
                shdr->sh_size, shdr->sh_flags );
	lst.push_back( sle );
  }
 
  rc = do_list_selection( w, COLOR_PAIR(3), COLOR_PAIR(6), lst );
  offs = -1;

  if ( rc != -1 && lst.size() ) {
  	scn = 0;
  	for ( i = 0; i <= rc; i++ )
		scn = elf_nextscn( elf, scn );
	shdr = elf32_getshdr( scn );
	offs = shdr->sh_offset;
  }
  elf_end( elf );

  delwin( w );
  pop_keyhints();

  return offs;
}

int do_search( int offs0, unsigned char *pat, int patlen, bool forw ) {
  int i, offs = offs0;
  unsigned char *buf, *p;

  buf = (unsigned char*) malloc( patlen );
  for ( i = 0; i < patlen; i++ ) {
  	p = get_data( offs + i );
	if ( !p ) {
		free( buf );
		return -1;
	}
	buf[i] = *p;
  }

  for (;;) {
  	if ( !memcmp( pat, buf, patlen ) ) break;

	if ( forw ) {
		memmove( buf, buf+1, patlen-1 );
		p = get_data( offs + patlen );
	}
	else {
		memmove( buf+1, buf, patlen-1 );
		p = get_data( offs-1 );
	}

	if ( !p ) {
		free( buf );
		return -1;
	}
	
	if ( forw ) {
		buf[patlen-1] = *p;
		offs++;
	}
	else {
		*buf = *p;
		offs--;
	}
  }

  free( buf );
  return offs;
}

void do_showabout() {
  WINDOW *win;
  int c;

  push_keyhints();
  clear_keyhints();
  win = newwin( 6, 51, (LINES-6)/2, (COLS-51)/2 );
  box( win, ACS_VLINE, ACS_HLINE );
  wbkgd( win, COLOR_PAIR(4) );
  mvwprintw( win, 1, 1, "Disassembler v1.0" );
  mvwprintw( win, 2, 1, "(c) 2003-2006 Copyright Vad Rulezz <vr5@narod.ru>" );
  mvwprintw( win, 3, 1, "Homepage: http://vr5.narod.ru/disasm" );
  mvwprintw( win, 4, 1, "Licensed under GPL v2 or later" );
  wrefresh( win );
  move( 0, COLS-1 );
  refresh();
  for (;;) {
	c = wgetch( win );
	if ( c != ('L' & 0x1f) ) break;
	clearok( stdscr, true );
	refresh();
  }
  delwin( win );
  pop_keyhints();
}

void do_showhex( int offs ) {
  int i, j;
  unsigned char *p;

  offs &= ~0xf;

  attrset( COLOR_PAIR(2) | A_BOLD );
  for ( i = 0; i < LINES-2; i++ ) {
  	mvprintw( i+1, 0, "%08X  ", offs );
	for ( j = 0; j < 8; j++ ) {
		p = get_data( offs + j );
		if ( p ) printw( "%02X ", *p );
		else addstr( "   " );
	}
	addstr( "- " );
	for ( j = 0; j < 8; j++ ) {
		p = get_data( offs + 8 + j );
		if ( p ) printw( "%02X ", *p );
		else addstr( "   " );
	}
	addch( ' ' );
	for ( j = 0; j < 16; j++ ) {
		p = get_data( offs + j );
		if ( !p ) {
			addch( ' ' );
			continue;
		}
		if ( isprint( *p ) && *p != 0x9b )
			addch( *p );
		else
			addch( '.' );
	}
	offs += 16;
  }
}

void do_showtext( int offs ) {
  int i, j, rc;
  unsigned char *p;
#ifdef USE_WIDEC
  wchar_t wc[2];
  cchar_t cc;
  char mbs[MB_CUR_MAX+1];

  wc[1] = 0;
  setlocale( LC_CTYPE, mb_locale );
#endif
 
  attrset( COLOR_PAIR(2) | A_BOLD );
  move( 1, 0 );
  for ( i = 0; i < LINES-2; i++ )
  	for ( j = 0; j < COLS; j++ ) {
		p = get_data( offs );
		if ( !p ) {
			addch( ' ' );
			offs++;
			continue;
		}
#ifdef USE_WIDEC
		rc = mbtowc( &wc[0], (char*)p, 1 );
		if ( rc != -1 && iswprint( wc[0] ) )
			addwstr( wc );
#else
		if ( *p != 0x9b && isprint( *p ) )
			addch( *p );
#endif
		else addch( '.' );
		offs++;
	}
#ifdef USE_WIDEC
  setlocale( LC_CTYPE, "" );
#endif
  refresh();
}

void do_highlight_text( int offs, int patoffs, int patlen ) {
  int i;
  unsigned char *p;

  if ( patoffs == -1 ||
       patoffs + patlen < offs ||
       patoffs > offs + (LINES-2)*COLS ) return;
  if ( patoffs < offs ) {
  	patlen -= offs - patoffs;
  	patoffs = offs;
  }
  if ( patoffs + patlen >= offs + (LINES-2)*COLS )
  	patlen = offs + (LINES-2)*COLS - patoffs;
  attrset( COLOR_PAIR(3) | A_REVERSE );
  move( (patoffs-offs)/COLS+1, (patoffs-offs)%COLS );
  for ( i = 0; i < patlen; i++ ) {
  	p = get_data( patoffs + i );
	if ( !p ) {
		addch( ' ' );
		continue;
	}
  	if ( *p != 0x9b && isprint(*p) ) addch( *p ); else addch( '.' );
  }
  move( 0, COLS-1 );
}
		
void do_highlight_hex( int offs, int patoffs, int patlen ) {
  unsigned char *p;
  int i, line, pos, poffs;

  if ( patoffs == -1 ||
       patoffs + patlen < offs ||
       patoffs > offs + (LINES-2)*16 ) return;
  if ( patoffs < offs ) {
  	patlen -= offs - patoffs;
  	patoffs = offs;
  }
  if ( patoffs + patlen >= offs + (LINES-2)*16 )
  	patlen = offs + (LINES-2)*16 - patoffs;
  attrset( COLOR_PAIR(3) | A_REVERSE );
  for ( i = 0; i < patlen; i++ ) {
  	line = (patoffs + i - offs) / 16 + 1;
	pos = (patoffs + i) & 0xf;
  	p = get_data( patoffs + i );
	if ( pos > 7 ) poffs = 12; else poffs = 10;
	mvprintw( line, (pos * 3) + poffs, "%02X", *p );
	if ( i != patlen - 1 ) {
		if ( pos != 15 ) addch( ' ' );
		if ( pos == 7 ) addstr( "- " );
	}

	if ( *p != 0x9b && isprint( *p ) ) mvaddch( line, pos + 61, *p );
	else mvaddch( line, pos + 61, '.' );
  }

  move( 0, COLS-1 );
}

void do_highlight_disasm( int offs, int patoffs, int patlen ) {
  int sz, i, j;
  struct instr insn;
  unsigned char *p;

  if ( patoffs == -1 ||
       patoffs + patlen < offs ) return;
  if ( patoffs < offs ) {
  	patlen -= offs - patoffs;
  	patoffs = offs;
  }
  attrset( COLOR_PAIR(3) | A_REVERSE );
  for ( i = 0; i < LINES-2; i++ ) {
  	if ( offs > patoffs + patlen ) break;
	p = get_data( offs );
	sz = disassemble_address_wrapper( offs, (char*)p, fsize - offs, &insn );
	if ( !sz ) sz = 1;
	for ( j = 0; j < sz; j++ )
		if ( offs + j >= patoffs && offs + j < patoffs + patlen )
			mvprintw( i + 1, j * 2 + 9, "%02X", p[j] );
	offs += sz;
  }
  move( 0, COLS-1 );
}

bool quit_requested;

void edit_hexcode( int offs ) {
  bool hi_nibble, ascii_mode;
  char istr[2];
  unsigned char *buf, *p;
  int c, val, val2, coffs, x, y, i, xtbl[16];
  int asc_pos;

  x = 10;
  for ( i = 0; i < 16; i++ ) {
  	xtbl[i] = x;
	x += 3;
	if ( i == 7 ) x += 2;
  }
  asc_pos = x + 1;

  buf = (unsigned char*) malloc( 16 * LINES-2 );
  for ( i = 0; i < 16 * LINES-2; i++ ) {
  	p = get_data( offs + i );
	if ( p ) buf[i] = *p; else buf[i] = 0;
  }

  push_keyhints();
  clear_keyhints();
  set_keyhint( 9, "Save" );
  set_keyhint( 10, "Quit" );

  hi_nibble = true;
  ascii_mode = false;
  coffs = 0;
  for (;;) {
  	y = coffs/16 + 1;
	if ( ascii_mode )
		x = asc_pos + coffs % 16;
	else {
		x = xtbl[coffs % 16];
		if ( !hi_nibble ) x++;
	}
	move( y, x );

get_next_key:
	c = getch();
	if ( c == KEY_RIGHT ) {
move_to_right:
		if ( ascii_mode ) {
			if ( coffs != (LINES-2) * 16 - 1 ) coffs++;
		}
		else {
			if ( !hi_nibble ) {
				if ( coffs == (LINES-2) * 16 - 1 )
					goto get_next_key;
				coffs++;
			}
			hi_nibble = !hi_nibble;
		}	
	}
	else if ( c == KEY_LEFT ) {
		if ( ascii_mode ) {
			if ( coffs > 0 ) coffs--;
		}
		else {
			if ( hi_nibble ) {
				if ( !coffs ) goto get_next_key;
				coffs--;
			}
			hi_nibble = !hi_nibble;
		}
	}
	else if ( c == KEY_DOWN ) {
		if ( coffs + 16 >= (LINES-2) * 16 ) goto get_next_key;
		coffs += 16;
	}
	else if ( c == KEY_UP ) {
		if ( coffs < 16 ) goto get_next_key;
		coffs -= 16;
	}
	else if ( c == '\t' ) {
		ascii_mode = !ascii_mode;
		if ( !ascii_mode ) hi_nibble = true;
	}
	else if ( c == KEY_BACKSPACE ) {
		if ( coffs == 0 ) goto get_next_key;

		if ( ascii_mode ) {
			coffs--;
			val = *get_data( offs + coffs );
			buf[coffs] = val;
			attrset( COLOR_PAIR(2) | A_BOLD );
			mvprintw( coffs / 16 + 1, xtbl[coffs % 16], "%02X", val );
			move( coffs / 16 + 1, asc_pos + coffs % 16 );
			if ( isprint(val) && val != 0x9b ) addch( val ); else addch( '.' );
		}
		else {
			if ( hi_nibble ) coffs--;
			hi_nibble = !hi_nibble;
			val = *get_data( offs + coffs );
			val2 = buf[coffs];
			attrset( COLOR_PAIR(2) | A_BOLD );
			if ( hi_nibble ) {
				buf[coffs] = val2 & 0xf | val & 0xf0;
				mvprintw( coffs / 16 + 1, xtbl[coffs % 16], "%X", val >> 4 );
			}
			else {
				buf[coffs] = val2 & 0xf0 | val & 0xf;
				mvprintw( coffs / 16 + 1, xtbl[coffs % 16] + 1, "%X", val & 0xf );
			}
			if ( buf[coffs] != val ) attrset( COLOR_PAIR(5) | A_BOLD );
			else attrset( COLOR_PAIR(2) | A_BOLD );
			move( coffs / 16 + 1, asc_pos + coffs % 16 );
			if ( isprint(buf[coffs]) && buf[coffs] != 0x9b )
				addch( buf[coffs] );
			else addch( '.' );
		}
	}
	else if ( c == KEY_F(9) ) {
	}
	else if ( c == KEY_F(10) ) {
		quit_requested = true;
		break;
	}
	else if ( c == ('L' & 0x1f) ) {
		clearok( stdscr, true );
		refresh();
	}
	else if ( c == 033 ) break;
	else if ( ascii_mode && c == (c & 0xff) ) {
		val2 = *get_data( offs + coffs );

		buf[coffs] = c;
		if ( val2 != c ) attrset( COLOR_PAIR(5) | A_BOLD );
		else attrset( COLOR_PAIR(2) | A_BOLD );

		if ( isprint( c ) && c != 0x9b ) addch( c );
		else addch( '.' );

		y = coffs / 16 + 1;
		x = xtbl[coffs % 16];
		move( y, x );
		if ( (val2 & 0xf0) != (c & 0xf0) )
			attrset( COLOR_PAIR(5) | A_BOLD );
		else attrset( COLOR_PAIR(2) | A_BOLD );
		printw( "%X", c >> 4 );
		if ( (val2 & 0xf) != (c & 0xf) )
			attrset( COLOR_PAIR(5) | A_BOLD );
		else attrset( COLOR_PAIR(2) | A_BOLD );
		printw( "%X", c & 0xf );

		goto move_to_right;
	}
	else if ( !ascii_mode && c == (c & 0xff) && isxdigit(c) ) {
		bool change = false;

		istr[0] = c;
		istr[1] = 0;
		val = strtol( istr, 0, 16 );
		val2 = *get_data( offs + coffs );
	
		p = &buf[coffs];
		if ( hi_nibble ) {
			val = val << 4 | (*p & 0xf);
			if ( (val & 0xf0) != (val2 & 0xf0) ) change = true;
		}
		else {
			val |= *p & 0xf0;
			if ( (val & 0xf) != (val2 & 0xf) ) change = true;
		}
		*p = val;

		if ( change ) attrset( COLOR_PAIR(5) | A_BOLD );
		else attrset( COLOR_PAIR(2) | A_BOLD );
	
		addch( toupper(c) );

		if ( val == val2 ) attrset( COLOR_PAIR(2) | A_BOLD );
		else attrset( COLOR_PAIR(5) | A_BOLD );
		move( coffs / 16 + 1, asc_pos + coffs % 16 );
		if ( isprint(buf[coffs]) && buf[coffs] != 0x9b )
			addch( buf[coffs] );
		else addch( '.' );

		goto move_to_right;
	}
  }
  free( buf );

  pop_keyhints();
  move( 0, 0 );
}

struct edit_disasm_state {
  int offs, line, pos, *offstbl;
  int bufsz;
  unsigned char *buf;
  bool hi_nibble;
};

void update_offstbl( edit_disasm_state& st ) {
  int i, coffs, sz;
  char *p;
  struct instr insn;
  
  coffs = st.offstbl[st.line];
  for ( i = st.line; i < LINES - 2; i++ ) {
  	st.offstbl[i] = coffs;
	if ( coffs >= st.bufsz - 20 ) {
		st.bufsz += 512;
		st.buf = (unsigned char*) realloc( st.buf, st.bufsz );
	}
  	sz = disassemble_address_wrapper( st.offs + coffs, (char*)&st.buf[coffs], st.bufsz - coffs, &insn );
	if ( !sz ) sz++;
	coffs += sz;
  }
  st.offstbl[i] = coffs;
}

void print_hexbyte_upd( int val, int val2 ) {
  if ( (val & 0xf0) != (val2 & 0xf0) ) {
	attrset( COLOR_PAIR(5) | A_BOLD );
	printw( "%X", val2 >> 4 );
	attrset( COLOR_PAIR(2) | A_BOLD );
  }
  else printw( "%X", val2 >> 4 );
  if ( (val & 0xf) != (val2 & 0xf) ) {
	attrset( COLOR_PAIR(5) | A_BOLD );
	printw( "%X", val2 & 0xf );
	attrset( COLOR_PAIR(2) | A_BOLD );
  }
  else printw( "%X", val2 & 0xf );
}

void update_bview( edit_disasm_state& st ) {
  int i, j, offs, val, val2, y, x, sz;
  unsigned char *p, *p2;
  struct instr insn;

  attrset( COLOR_PAIR(2) | A_BOLD );
  for ( i = st.line; i < LINES-2; i++ ) {
  	memset( &insn, 0, sizeof insn );
	offs = st.offstbl[i];
	p = &st.buf[offs];
  	sz = disassemble_address_wrapper( st.offs + offs, (char*)p, st.bufsz - offs, &insn );
	mvprintw( i + 1, 0, "%08X ", st.offs + offs );
	if ( sz ) {
		for ( j = 0; j < 8; j++ )
			if ( j < sz ) {
				val = *get_data( st.offs + offs + j );
				val2 = st.buf[offs + j];
				print_hexbyte_upd( val, val2 );
			}
			else printw( "  " );
		printw( " %s", insn.mnemonic );
		if ( strlen(insn.dest) ) printw( " %s", insn.dest );
		if ( strlen(insn.src) ) printw( ", %s", insn.src );
		if ( strlen(insn.aux) ) printw( ", %s", insn.aux );
	}
	else {
		val = *get_data( st.offs + offs );
		val2 = st.buf[offs];
		print_hexbyte_upd( val, val2 );
		addstr( "               (bad)" );
	}
	getyx( stdscr, y, x );
	for ( j = x; j < COLS; j++ ) addch( ' ' );
  }
}

int do_translate_asm( unsigned char *codebuf, int cblen );

void edit_disasm( int offs ) {
  int c, i, val, val2, sz, coffs;
  unsigned char *p;
  char istr[2];
  bool change;
  struct instr insn;
  edit_disasm_state st;

  st.offstbl = (int*) malloc( (LINES - 1) * sizeof(int) );
  st.offs = offs;
  coffs = offs;
  for ( i = 0; i < LINES - 2; i++ ) {
  	st.offstbl[i] = coffs - offs;
  	p = get_data( coffs );
  	sz = disassemble_address_wrapper( coffs, (char*)p, fsize - coffs, &insn );
	if ( !sz ) sz++;
	coffs += sz;
  }
  st.offstbl[i] = coffs - offs;

  push_keyhints();
  clear_keyhints();
  set_keyhint( 7, "Asm" );
  set_keyhint( 9, "Save" );
  set_keyhint( 10, "Quit" );
 
  st.bufsz = (coffs - offs + 511) & ~511;
  st.buf = (unsigned char*) malloc( st.bufsz );
  for ( i = 0; i < st.bufsz; i++ ) {
  	p = get_data( offs + i );
	st.buf[i] = *p;
  }

  coffs = 0;
  st.hi_nibble = true;
  st.line = 0;
  st.pos = 0;
  for (;;) {
  	move( st.line+1, st.pos+9 );
get_next_key:
	c = getch();
	if ( c == KEY_DOWN ) {
move_down:
		if ( st.line+1 >= LINES-2 ) goto get_next_key;
		st.line++;
		if ( st.pos >= (st.offstbl[st.line+1] - st.offstbl[st.line]) * 2 ) {
			st.pos = (st.offstbl[st.line+1] - st.offstbl[st.line]) * 2 - 1;
			st.hi_nibble = false;
		}
	}
	else if ( c == KEY_UP ) {
		if ( !st.line ) goto get_next_key;
		st.line--;
		if ( st.pos >= (st.offstbl[st.line+1] - st.offstbl[st.line]) * 2 ) {
			st.pos = (st.offstbl[st.line+1] - st.offstbl[st.line]) * 2 - 1;
			st.hi_nibble = false;
		}
	}
	else if ( c == KEY_RIGHT ) {
move_to_right:
		if ( st.pos+1 < (st.offstbl[st.line+1] - st.offstbl[st.line]) * 2) {
			st.pos++;
			st.hi_nibble = !st.hi_nibble;
		}
		else
			if ( st.line+1 < LINES-2 ) {
				st.line++;
				st.pos = 0;
				st.hi_nibble = true;
			}
	}
	else if ( c == KEY_LEFT ) {
		if ( st.pos ) {
			st.pos--;
			st.hi_nibble = !st.hi_nibble;
		}
		else
			if ( st.line ) {
				st.line--;
				st.pos = (st.offstbl[st.line+1] - st.offstbl[st.line]) * 2 - 1;
				st.hi_nibble = false;
			}
	}
	else if ( c == KEY_BACKSPACE ) {
		if ( st.pos ) st.pos--;
		else {
			if ( st.line ) {
				st.line--;
				st.pos = (st.offstbl[st.line+1] - st.offstbl[st.line]) * 2 - 1;
			}
			else goto get_next_key;
		}
		st.hi_nibble = !st.hi_nibble;
	
		coffs = st.offstbl[st.line] + st.pos/2;
		val = st.buf[coffs];
		val2 = *get_data( offs + coffs );
		if ( st.hi_nibble )
			st.buf[coffs] = val & 0xf | val2 & 0xf0;
		else
			st.buf[coffs] = val & 0xf0 | val2 & 0xf;

		update_offstbl( st );
		update_bview( st );
	}
	else if ( c == (c & 0xff) && isxdigit(c) ) {
		istr[0] = c;
		istr[1] = 0;
		val = strtol( istr, 0, 16 );

		coffs = st.offstbl[st.line] + st.pos/2;
		val2 = *get_data( offs + coffs );
		if ( st.hi_nibble )
			st.buf[coffs] = (st.buf[coffs] & 0xf) | (val << 4);
		else
			st.buf[coffs] = (st.buf[coffs] & 0xf0) | val;

		update_offstbl( st );
		update_bview( st );
		goto move_to_right;
	}
	else if ( c == KEY_F(7) ) {
		coffs = st.offstbl[st.line] + st.pos/2;
		do_translate_asm( &st.buf[coffs], st.bufsz - coffs );
		update_offstbl( st );
		update_bview( st );
		goto move_down;
	}
	else if ( c == 033 ) break;
	else if ( c == KEY_F(10) ) {
		quit_requested = true;
		break;
	}
  }

  pop_keyhints();

  free( st.buf );
  free( st.offstbl );
}

void do_disasm( int offs ) {
  int sz, i, j, x, y;
  unsigned char *p;
  struct instr insn;

  attrset( COLOR_PAIR(2) | A_BOLD );
  for ( i = 0; i < LINES-2; i++ ) {
  	memset( &insn, 0, sizeof insn );
	p = get_data( offs );
	if ( !p ) break;
  	sz = disassemble_address_wrapper( offs, (char*)p, fsize - offs, &insn );
	if ( sz ) {
		mvprintw( i + 1, 0, "%08X ", offs );
		for ( j = 0; j < 8; j++ )
			if ( j < sz ) printw( "%02X", p[j] );
			else printw( "  " );
		printw( " %s", insn.mnemonic );
		if ( strlen(insn.dest) ) printw( " %s", insn.dest );
		if ( strlen(insn.src) ) printw( ", %s", insn.src );
		if ( strlen(insn.aux) ) printw( ", %s", insn.aux );
		offs += sz;
	}
	else {
		mvprintw( i + 1, 0, "%08X %02X               (bad)",
			offs, ((unsigned char*)p)[0] );
		offs++;
	}
	getyx( stdscr, y, x );
	for ( j = x; j < COLS; j++ ) addch( ' ' );
  }
}

void try_scroll_down( int& offs, int lines=1 ) {
  struct instr insn;
  unsigned char *p;
  int sz;

  while ( lines ) {
	p = get_data( offs );
	sz = disassemble_address_wrapper( offs, (char*)p, fsize - offs, &insn );
	if ( !sz && fsize - offs ) sz++;
	offs += sz;
	lines--;
  }
}

void try_scroll_up( int& offs, int lines=1 ) {
  int sz, i, best, offs2;
  struct instr insn;
  unsigned char *p;

  while ( offs && lines ) {
  	best = 1;

  	for ( i = 2; i < 9; i++ ) {
		offs2 = offs - i;
		if ( offs2 < 0 ) break;

		p = get_data( offs2 );
		sz = disassemble_address_wrapper( offs2, (char*)p, fsize - offs2, &insn );
		if ( sz == i ) best = i;
	}

	offs -= best;
	lines--;
  }
}

enum disasm_mode_type {
  DISASM_MODE,
  HEX_MODE,
  TEXT_MODE
};

disasm_mode_type disasm_mode;

struct string_ent_state {
  int y, x, idx;
  bool hi_nibble;
  unsigned char *buf;
  int bsz;
  WINDOW *win;
};

void enter_offset( int& offs );
void init_cstr_entry( string_ent_state *st, WINDOW *win,
	unsigned char *buf, int bsz );
void init_xstr_entry( string_ent_state *st, WINDOW *win,
	unsigned char *buf, int bsz );
int enter_cstring( string_ent_state *st, int& c, bool single_char );
int enter_xstring( string_ent_state *st, int& c, bool single_char );
void update_cstring( string_ent_state *st );
void update_xstring( string_ent_state *st );

bool do_asm_str_x86( char *str );
extern unsigned char asmbuf_x86[50], *asmbuf_ptr_x86;
bool do_asm_str_arm( char *str );
extern unsigned char asmbuf_arm[50], *asmbuf_ptr_arm;

bool do_asm_str( char *str ) {
  if ( x86_mode )
  	return do_asm_str_x86( str );
  
  return do_asm_str_arm( str );
}

unsigned char *get_asm_data( int& datalen ) {
  if ( x86_mode ) {
  	datalen = asmbuf_ptr_x86 - asmbuf_x86;
	return asmbuf_x86;
  }

  datalen = asmbuf_ptr_arm - asmbuf_arm;
  return asmbuf_arm;
}

int do_translate_asm( unsigned char *codebuf, int cblen ) {
  WINDOW *w;
  unsigned char buf[40];
  string_ent_state st;
  int c;

  w = newwin( 3, 40, (LINES-3)/2, (COLS-40)/2 );
  keypad( w, TRUE );
  box( w, ACS_VLINE, ACS_HLINE );
  wbkgd( w, COLOR_PAIR(4) );
  mvwprintw( w, 0, 2, "x86 assembler" );
  
  wmove( w, 1, 1 );
  init_cstr_entry( &st, w, buf, 38 );
  for (;;) {
	enter_cstring( &st, c, false );
	if ( c == 033 || c == '\n' ) break;
  }
  delwin( w );
  if ( c == 033 ) return -1;

  buf[st.idx] = 0;
  if ( !do_asm_str( (char*) buf ) ) return -1;
  
  unsigned char *asmbuf;
  int asmlen;
  asmbuf = get_asm_data( asmlen );

  if ( asmlen < cblen ) cblen = asmlen;
  memcpy( codebuf, asmbuf, cblen );
  
  return cblen;
}

bool do_enter_search( unsigned char **spat, int *lpat, bool& forward ) {
  unsigned char *buf;
  WINDOW *win;
  int n, c, done, rc;
  string_ent_state xst, cst;
  bool ascii = false;

  push_keyhints();
  clear_keyhints();
  set_keyhint( 2, "Dir" );
  set_keyhint( 7, "Asm" );
  refresh();
 
  if ( !spat ) return false;
  if ( *spat )
  	buf = *spat;
  else
  	buf = *spat = (unsigned char*) malloc( 50 );
 
  win = newwin( 6, 50, (LINES-6)/2, (COLS-50)/2 );
  keypad( win, TRUE );
  box( win, ACS_VLINE, ACS_HLINE );
  wbkgd( win, COLOR_PAIR(4) );
  mvwprintw( win, 0, 2, "Search/" );
  waddstr( win, forward? "Forward": "Backward" );
  mvwprintw( win, 1, 1, "Ascii" );
  mvwprintw( win, 3, 1, "Hex" );
  wmove( win, 2, 1 );
  init_xstr_entry( &cst, win, buf, 16 );
  cst.idx = *lpat;
  update_cstring( &cst );
  wmove( win, 4, 1 );
  init_xstr_entry( &xst, win, buf, 16 );
  xst.idx = *lpat;
  update_xstring( &xst );
  for (;;) {
  	if ( ascii ) {
		done = enter_cstring( &cst, c, true );
		xst.idx = cst.idx;
		xst.hi_nibble = true;
		update_xstring( &xst );
		if ( done ) break;
	}
	else {
		done = enter_xstring( &xst, c, true );
		cst.idx = xst.idx;
		if ( !xst.hi_nibble ) cst.idx++;
		update_cstring( &cst );
		if ( done ) break;
	}

	if ( c == KEY_UP || c == KEY_DOWN || c == '\t' )
		ascii = !ascii;
	else if ( c == KEY_F(2) ) {
		forward = !forward;
		if ( forward ) {
			mvwaddstr( win, 0, 9, "Forward" );
			waddch( win, ACS_HLINE );
		}
		else
			mvwaddstr( win, 0, 9, "Backward" );
	}
	else if ( c == KEY_F(7) ) {
		rc = do_translate_asm( buf + cst.idx, 16 - cst.idx );
		touchwin( win );
		wrefresh( win );
		if ( rc != -1 ) {
			cst.idx += rc;
			xst.idx += rc;
			update_cstring( &cst );
			update_xstring( &xst );
		}
	}
	else if ( c == ('L' & 0x1f) ) {
		clearok( stdscr, true );
		refresh();
	}
  }
  delwin( win );

  pop_keyhints();

  *lpat = cst.idx;
  if ( done == 1 ) return true; else return false;
}

void interactive_asm();

void disasm_loop( int& offs ) {
  int c, rc, patoffs = -1, lpat = 0;
  unsigned char *spat = 0;
  bool forw = true;

  push_keyhints();
  if ( !view_objtbl ) set_keyhint( 6, "      " );
  set_keyhint( 2, "IntAsm" );
  set_keyhint( 8, x86_mode? "ArmIsa": "x86Isa" );

  do_disasm( offs );

  while ( !quit_requested ) {
	c = getch();
	switch (c) {
	case KEY_DOWN:
		try_scroll_down( offs );
		do_disasm( offs );
		do_highlight_disasm( offs, patoffs, lpat );
		break;
	case KEY_UP:
		try_scroll_up( offs );
		do_disasm( offs );
		do_highlight_disasm( offs, patoffs, lpat );
		break;
	case KEY_NPAGE:
		try_scroll_down( offs, LINES-2 );
		do_disasm( offs );
		do_highlight_disasm( offs, patoffs, lpat );
		break;
	case KEY_PPAGE:
		try_scroll_up( offs, LINES-2 );
		do_disasm( offs );
		do_highlight_disasm( offs, patoffs, lpat );
		break;
	case KEY_HOME:
		offs = 0;
		do_disasm( offs );
		do_highlight_disasm( offs, patoffs, lpat );
		break;
	case KEY_F(1):
		do_showabout();
		do_disasm( offs );
		do_highlight_disasm( offs, patoffs, lpat );
		break;
	case KEY_F(2):
		interactive_asm();
		do_disasm( offs );
		break;
	case KEY_F(3):
		edit_disasm( offs );
		do_disasm( offs );
		do_highlight_disasm( offs, patoffs, lpat );
		break;
	case KEY_F(4):
	case '\n':
		disasm_mode = HEX_MODE;
		goto leave_mode;
	case KEY_F(5):
		enter_offset( offs );
		do_disasm( offs );
		do_highlight_disasm( offs, patoffs, lpat );
		break;
	case KEY_F(6):
		if ( view_objtbl ) {
			rc = view_objtbl();
			if ( rc != -1 ) offs = rc;
		}
		do_disasm( offs );
		do_highlight_disasm( offs, patoffs, lpat );
		break;
	case KEY_F(8):
		x86_mode = !x86_mode;
  		set_keyhint( 8, x86_mode? "ArmIsa": "x86Isa" );
		do_disasm( offs );
		break;
	case KEY_F(17):
		if ( patoffs != -1 && patoffs < fsize-1 ) {
			int srch_from;
	
			srch_from = forw? patoffs+1: patoffs-1; 
			patoffs = do_search( srch_from, spat, lpat, forw );
			goto search_done;
		}
		break;
	case KEY_F(7):
		if ( do_enter_search( &spat, &lpat, forw ) ) {
			patoffs = do_search( offs, spat, lpat, forw );
search_done:
			if ( patoffs != -1 ) offs = patoffs;
		}
		do_disasm( offs );
		do_highlight_disasm( offs, patoffs, lpat );
		break;
	case 'L' & 0x1f:
		clearok( stdscr, true );
		refresh();
		break;
	case 033:
	case KEY_F(10):
		quit_requested = true;
		break;;
	}
  }

leave_mode:
  pop_keyhints();
}

void hex_loop( int& offs ) {
  int c, rc, lpat = 0, patoffs = -1;
  unsigned char *spat = 0;
  bool forw = true;

  push_keyhints();
  if ( !view_objtbl ) set_keyhint( 6, "      " );

  offs &= ~0xf;
  do_showhex( offs );

  while ( !quit_requested ) {
	c = getch();
	switch (c) {
	case KEY_DOWN:
		if ( offs + (LINES-2) * 16 < fsize ) offs += 16;
//		else {
//			if ( fsize > (LINES-2) * 16 )
//				offs = fsize - (LINES-2) * 16;
//			else offs = 0;
//		}
		do_showhex( offs );
		do_highlight_hex( offs, patoffs, lpat );
		break;
	case KEY_UP:
		if ( offs >= 16 ) {
			offs -= 16;
			do_showhex( offs );
			do_highlight_hex( offs, patoffs, lpat );
		}
		else offs = 0;
		break;
	case KEY_NPAGE:
		if ( offs + (LINES-2) * 16 * 2 < fsize ) offs += (LINES-2) * 16;
		else {
			if ( fsize > (LINES-2) * 16 ) {
				offs = (fsize - (LINES-2) * 16) & ~0xf;
			}
			else offs = 0;
		}

		do_showhex( offs );
		do_highlight_hex( offs, patoffs, lpat );
		break;
	case KEY_PPAGE:
		if ( offs > (LINES-2) * 16 ) offs -= (LINES-2) * 16;
		else offs = 0;

		do_showhex( offs );
		do_highlight_hex( offs, patoffs, lpat );
		break;
	case KEY_HOME:
		offs = 0;
		do_showhex( offs );
		do_highlight_hex( offs, patoffs, lpat );
		break;
	case KEY_END:
		offs = (fsize - (LINES-2) * 16 & ~0xf);
		if ( offs < 0 ) offs = 0;
		do_showhex( offs );
		do_highlight_hex( offs, patoffs, lpat );
		break;
	case KEY_F(1):
		do_showabout();
		do_showhex( offs );
		do_highlight_hex( offs, patoffs, lpat );
		break;
	case KEY_F(3):
		edit_hexcode( offs );
		do_showhex( offs );
		do_highlight_hex( offs, patoffs, lpat );
		break;
	case KEY_F(4):
	case '\n':
		disasm_mode = TEXT_MODE;
		goto leave_mode;
	case KEY_F(5):
		enter_offset( offs );
		do_showhex( offs );
		do_highlight_hex( offs, patoffs, lpat );
		break;
	case KEY_F(6):
		if ( view_objtbl ) {
			rc = view_objtbl();
			if ( rc != -1 ) offs = rc;
		}
		do_showhex( offs );
		do_highlight_hex( offs, patoffs, lpat );
		break;
	case KEY_F(17):
		if ( patoffs != -1 && patoffs < fsize-1 ) {
			int srch_from;
	
			srch_from = forw? patoffs+1: patoffs-1; 
			patoffs = do_search( srch_from, spat, lpat, forw );
			goto search_done;
		}
		break;
	case KEY_F(7):
		if ( do_enter_search( &spat, &lpat, forw ) ) {
			patoffs = do_search( offs, spat, lpat, forw );
search_done:	
			if ( patoffs != -1 ) {
				if ( patoffs < offs ||
				     patoffs > offs + (LINES-2)*16 ) {
					offs = patoffs;
					offs -= offs & 0xf;
				}
			}
		}
		do_showhex( offs );
		do_highlight_hex( offs, patoffs, lpat );
		break;
	case 'L' & 0x1f:
		clearok( stdscr, true );
		refresh();
		break;
	case 033:
	case KEY_F(10):
		quit_requested = true;
		break;
	}
  }
leave_mode:
  pop_keyhints();
}

void text_loop( int& offs ) {
  int c, patoffs = -1, lpat = 0;
  unsigned char *spat = 0;
  bool forw = true;

  push_keyhints();
  set_keyhint( 3, "      " );
  set_keyhint( 5, "      " );
  set_keyhint( 6, "      " );

  do_showtext( offs );

  while ( !quit_requested ) {
  	c = getch();
	switch ( c ) {
	case KEY_DOWN:
		offs += COLS;
		if ( offs + (LINES-2) * COLS > fsize ) {
			offs = fsize + COLS-1 - (LINES-2) * COLS;
			offs -= offs % COLS;
			if ( offs < 0 ) offs = 0;
		}
		do_showtext( offs );
		do_highlight_text( offs, patoffs, lpat );
		break;
	case KEY_UP:
		offs -= COLS;
		if ( offs < 0 ) offs = 0;
		do_showtext( offs );
		do_highlight_text( offs, patoffs, lpat );
		break;
	case KEY_NPAGE:
		offs += (LINES-2) * COLS;
		if ( offs + (LINES-2) * COLS > fsize ) {
			offs = fsize + COLS-1 - (LINES-2) * COLS;
			offs -= offs % COLS;
			if ( offs < 0 ) offs = 0;
		}
		do_showtext( offs );
		do_highlight_text( offs, patoffs, lpat );
		break;
	case KEY_PPAGE:
		offs -= (LINES-2) * COLS;
		if ( offs < 0 ) offs = 0;
		do_showtext( offs );
		do_highlight_text( offs, patoffs, lpat );
		break;
	case KEY_HOME:
		offs = 0;
		do_showtext( offs );
		do_highlight_text( offs, patoffs, lpat );
		break;
	case KEY_END:
		offs = fsize + COLS-1 - (LINES-2) * COLS;
		offs -= offs % COLS;
		if ( offs < 0 ) offs = 0;
		do_showtext( offs );
		do_highlight_text( offs, patoffs, lpat );
		break;
	case KEY_F(1):
		do_showabout();
		do_showtext( offs );
		do_highlight_text( offs, patoffs, lpat );
		break;
	case KEY_F(4):
	case '\n':
		disasm_mode = DISASM_MODE;
		goto leave_mode;
	case KEY_F(17):
		if ( patoffs != -1 && patoffs < fsize-1 ) {
			int srch_from;
	
			srch_from = forw? patoffs+1: patoffs-1; 
			patoffs = do_search( srch_from, spat, lpat, forw );
			goto search_done;
		}
		break;
	case KEY_F(7):
		if ( do_enter_search( &spat, &lpat, forw ) ) {
			patoffs = do_search( offs, spat, lpat, forw );
search_done:
			if ( patoffs != -1 ) {
				if ( patoffs < offs ||
				     patoffs > offs + (LINES-2)*COLS ) {
					offs = patoffs;
					offs -= offs % COLS;
				}
			}
		}
		do_showtext( offs );
		do_highlight_text( offs, patoffs, lpat );
		break;
	case 'L' & 0x1f:
		clearok( stdscr, true );
		refresh();
		break;
	case 033:
	case KEY_F(10):
		quit_requested = true;
		break;
	}
  }

leave_mode:
  pop_keyhints();
}

void enter_offset( int& offs ) {
  int i, c, req_offs;
  char buf[9], *p;

  push_keyhints();
  clear_keyhints();

  move( 1, 0 );
  attrset( COLOR_PAIR(2) | A_BOLD );
  for ( i = 0; i < 8; i++ ) addch( ACS_CKBOARD );

  move ( 1, 0 );
  p = buf;
  for (;;) {
  	c = getch();
	if ( c == (c & 0xff) && isxdigit(c) && p < buf + 8 ) {
		c = toupper(c);
		addch( c );
		*p++ = c;
	}
	else if ( c == KEY_BACKSPACE ) {
		p--;
		mvaddch( 1, p - buf, ACS_CKBOARD );
		move( 1, p - buf );
	}
	else if ( c == '\n' ) {
		*p = 0;
		break;
	}
	else if ( c == ('L' & 0x1f) ) {
		clearok( stdscr, true );
		refresh();
	}
	else if ( c == 033 )
		goto cleanup;
  }

  req_offs = strtol( buf, 0, 16 );
  if ( req_offs >= 0 && req_offs < fsize ) offs = req_offs;

cleanup:
  pop_keyhints();
}

void init_cstr_entry( string_ent_state *st, WINDOW *win,
	unsigned char *buf, int bsz ) {
  st->win = win;
  st->buf = buf;
  st->bsz = bsz;
  
  getyx( st->win, st->y, st->x );
  whline( st->win, ACS_CKBOARD, st->bsz - 1 );
  wmove( st->win, st->y, st->x );
  st->idx = 0;
}

void init_xstr_entry( string_ent_state *st, WINDOW *win,
	unsigned char *buf, int bsz ) {
  st->win = win;
  st->buf = buf;
  st->bsz = bsz;
  
  getyx( st->win, st->y, st->x );
  whline( st->win, ACS_CKBOARD, st->bsz * 3 - 1 );
  wmove( st->win, st->y, st->x );
  st->idx = 0;
  st->hi_nibble = true;
}

int enter_cstring( string_ent_state *st, int& c, bool single_char ) {
  wmove( st->win, st->y, st->x + st->idx );
  for (;;) {
  	c = wgetch( st->win );
	if ( c == '\n' ) return 1;
	else if ( c == 033 ) return 2;
	else if ( c == '\t' ) return 0;
	else if ( c == (c & 0xff) && st->idx < st->bsz ) {
		if ( c != 0x9b && isprint( c ) ) waddch( st->win, c );
		else waddch( st->win, '.' );
		st->buf[st->idx++] = c;
	}
	else if ( c == KEY_BACKSPACE && st->idx ) {
		st->idx--;
		mvwaddch( st->win, st->y, st->x + st->idx, ACS_CKBOARD );
		wmove( st->win, st->y, st->x + st->idx );
	}
	else break;
	if ( single_char ) break;
  }

  return 0;
}

void update_cstring( string_ent_state *st ) {
  int i, c;

  wmove( st->win, st->y, st->x );
  for ( i = 0; i < st->idx; i++ ) {
  	c = st->buf[i];
	if ( c == 0x9b || !isprint( c ) ) c = '.';
	waddch( st->win, c );
  }
  for (; i < st->bsz; i++ ) waddch( st->win, ACS_CKBOARD );
  wrefresh( st->win );
}

int enter_xstring( string_ent_state *st, int& c, bool single_char ) {
  int n, xpos;
  char s[2];

  if ( st->hi_nibble ) {
  	if ( st->idx ) xpos = st->x + st->idx * 3 - 1;
	else xpos = st->x;
  }
  else xpos = st->x + st->idx * 3 + 1;
  wmove( st->win, st->y, xpos );

  for (;;) {
  	c = wgetch( st->win );
	if ( c == (c & 0xff) && isxdigit(c) && st->idx < st->bsz ) {
		s[0] = c;
		s[1] = 0;
		n = strtol( s, 0, 16 );
		if ( st->hi_nibble ) {
			if ( st->idx ) waddch( st->win, ' ' );
			waddch( st->win, toupper(c) );
			st->buf[st->idx] = n << 4;
		}
		else {
			waddch( st->win, toupper(c) );
			n |= st->buf[st->idx] & 0xf0;
			st->buf[st->idx++] = n;
		}
		st->hi_nibble = !st->hi_nibble;
	}
	else if ( c == KEY_BACKSPACE ) {
		if ( st->hi_nibble ) {
			if ( !st->idx ) continue;
			mvwaddch( st->win, st->y, st->x + st->idx * 3 - 2,
				ACS_CKBOARD );
			wmove( st->win, st->y, st->x + st->idx * 3 - 2 );
			st->idx--;
			st->buf[st->idx] &= 0xf0;
		}
		else {
			if ( st->idx ) {
				mvwaddch( st->win, st->y,
					st->x + st->idx * 3 - 1, ACS_CKBOARD );
				waddch( st->win, ACS_CKBOARD );
				wmove( st->win, st->y,
					st->x + st->idx * 3 - 1 );
			}
			else {
				mvwaddch( st->win, st->y, st->x, ACS_CKBOARD );
				wmove( st->win, st->y, st->x );
			}
		}
		st->hi_nibble = !st->hi_nibble;
	}
	else if ( st->hi_nibble && c == '\n' ) return 1;
	else if ( c == 033 ) return 2;
	else break;
	if ( single_char ) break;
  }
  return 0;
}

void update_xstring( string_ent_state *st ) {
  int i, c;

  wmove( st->win, st->y, st->x );
  for ( i = 0; i < st->idx; i++ ) {
  	c = st->buf[i];
	if ( i ) waddch( st->win, ' ' );
	wprintw( st->win, "%02X", c );
  }
  for (; i < st->bsz; i++ ) {
  	waddch( st->win, ACS_CKBOARD );
  	waddch( st->win, ACS_CKBOARD );
  	if ( i != st->bsz-1 ) waddch( st->win, ACS_CKBOARD );
  }
  wrefresh( st->win );
}

void interactive_asm() {
  bool quit_requested = false;
  int xpos, ypos, c, minx, miny, maxy, maxx, i;
  char textbuf[50], *p;
  WINDOW *w;

  push_keyhints();
  clear_keyhints();
  set_keyhint( 10, "Exit" );
  refresh();

  w = newwin( 15, 51, (LINES-15)/2, (COLS-51)/2 );
  keypad( w, TRUE );
  box( w, ACS_VLINE, ACS_HLINE );
  wbkgd( w, COLOR_PAIR(6) );
  mvwprintw( w, 0, 2, "Interactive Assembler" );

  minx = 18;
  miny = 1;
  getmaxyx( w, maxy, maxx );
  maxx -= 2;
  maxy -= 2;

  xpos = minx;
  ypos = miny;
  wmove( w, ypos, xpos );

  while ( !quit_requested ) {
  	c = wgetch( w );
	switch ( c ) {
	case KEY_DOWN:
		if ( ypos < maxy ) {
			ypos++;
			wmove( w, ypos, xpos );
		}
		break;
	case KEY_UP:
		if ( ypos > miny ) {
			ypos--;
			wmove( w, ypos, xpos );
		}
		break;
	case KEY_RIGHT:
		if ( xpos < maxx ) {
			xpos++;
			wmove( w, ypos, xpos );
		}
		break;
	case KEY_LEFT:
		if ( xpos > minx ) {
			xpos--;
			wmove( w, ypos, xpos );
		}
		break;
	case KEY_HOME:
		xpos = minx;
		wmove( w, ypos, xpos );
		break;
	case KEY_END:
		xpos = maxx;
		wmove( w, ypos, xpos );
		break;
	case '\n':
	case KEY_ENTER:
		if ( ypos < maxy ) {
			xpos = minx;
			ypos++;
			wmove( w, ypos, xpos );
		}
		break;
	case KEY_BACKSPACE:
		if ( xpos > minx ) {
			xpos--;
			mvwaddch( w, ypos, xpos, ' ' );
			wmove( w, ypos, xpos );
		}
		break;
	case 'L' & 0x1f:
		clearok( stdscr, true );
		refresh();
		break;
	case 033:
	case KEY_F(10):
		quit_requested = true;
		break;
	default:
		if ( c == (c & 0xff) && c != 0x9b && isprint( c )  )
			mvwaddch( w, ypos, xpos, c );
		else
			mvwaddch( w, ypos, xpos, '.' );

		mvwinnstr( w, ypos, minx, textbuf, maxx - minx );

		p = textbuf;
		while ( *p == ' ' ) p++;
		memmove( textbuf, p, strlen(p) + 1 );
		p = textbuf + strlen( textbuf ) - 1;
		while ( p > textbuf && *p == ' ' ) {
			*p = 0;
			p--;
		}

		wmove( w, ypos, 1 );

		if ( do_asm_str( textbuf ) ) {
			unsigned char *asmbuf;
			int asmlen;

			asmbuf = get_asm_data( asmlen );
			for ( i = 0; i < asmlen && i < 8; i++ )
				wprintw( w, "%02X", asmbuf[i] );
		}
		else i = 0;

		for ( ; i < 8; i++ )
			waddstr( w, "  " );

		xpos++;
		if ( xpos > maxx ) {
			if ( ypos < maxy ) {
				xpos = minx;
				ypos++;
			}
			else
				xpos = maxx;
		}
		wmove( w, ypos, xpos );
	}
  }

leave_mode:
  delwin( w );
  pop_keyhints();
}

int main( int argc, char **argv ) {
  int offs;

  if ( argc < 2 ) {
  	puts( "Syntax: ./disasm <filename>" );
	return 1;
  }
  fd = open( argv[1], O_RDONLY );
  if ( fd == -1 ) {
  	perror( "disasm open" );
	return 1;
  }
  read( fd, buf, sizeof buf );
  fsize = lseek( fd, 0, SEEK_END );
  buf_offs = 0;

  if ( load_le_objtbl() )
  	view_objtbl = view_le_objtbl;
  else if ( load_pe_objtbl() )
  	view_objtbl = view_pe_objtbl;
  else if ( load_elf_objtbl() )
  	view_objtbl = view_elf_objtbl;

  setlocale( LC_CTYPE, "" );

  initscr();
  start_color();
  cbreak();
  noecho();
  disasm_init( argv[1] );
  keypad( stdscr, true );

  quit_requested = false;
  disasm_mode = DISASM_MODE;
  rel_op_callback = relative_op_callback;
  disassemble_init( 0, INTEL_SYNTAX );
  offs = 0;

  while ( !quit_requested ) {
  	if ( disasm_mode == DISASM_MODE )
		disasm_loop( offs );
	else if ( disasm_mode == HEX_MODE )
		hex_loop( offs );
	else
		text_loop( offs );
  }

do_exit: 
  erase();
  refresh();
  endwin();
}
